from enum import Enum
from functools import partial
from typing import Any, Callable, List, Optional, Union

import numpy as np
import pandas as pd
import scipy.stats as stats

from opti.parameter import Parameters
from opti.problem import Problem


def _add_noise_to_data(Y, noisifiers, outputs: Parameters):
    Y = np.atleast_2d(Y)
    noisy_Y = np.stack(
        [noisify(Y[:, col]) for noisify, col in zip(noisifiers, range(Y.shape[1]))],
        axis=1,
    )
    for col, output in enumerate(outputs):
        noisy_Y[:, col] = output.round(noisy_Y[:, col])
    return noisy_Y


def noisify_problem(
    problem: Problem,
    noisifiers: Union[Callable, List[Callable]],
    name: Optional[str] = None,
) -> Problem:
    """Creates a new problem that is based on the given one plus noise on the outputs.

    Args:
        problem: given problem where we will add noise
        noisifiers: function or list of functions that add noise to the outputs
        name: name of the new problem

    Returns: new problem with noise on the output
    """

    def noisy_f(X):
        nonlocal noisifiers
        if isinstance(noisifiers, Callable):
            noisifiers = [noisifiers] * problem.n_outputs
        Y = problem.f(X)

        return _add_noise_to_data(Y, noisifiers, problem.outputs)

    if problem.data is not None:
        noisy_Y = _add_noise_to_data(
            problem.data[problem.outputs.names].values, noisifiers, problem.outputs
        )
        data = pd.concat(
            [
                problem.data[problem.inputs.names],
                pd.DataFrame(
                    columns=problem.outputs.names,
                    index=problem.data.index,
                    data=noisy_Y,
                ),
            ],
            axis=1,
        )
    else:
        data = None

    return Problem(
        inputs=problem.inputs,
        outputs=problem.outputs,
        objectives=problem.objectives,
        constraints=problem.constraints,
        output_constraints=problem.output_constraints,
        f=noisy_f,
        data=data,
        name=problem.name if name is None else name,
    )


class NoiseType(Enum):
    ADDITIVE = 1
    MULTIPLICATIVE = 2


def noisify_problem_with_scipy_stats(
    problem: Problem,
    random_variables: Union[list, Any],
    rv_kwargs: Union[None, dict, List[dict]] = None,
    noise_types: Union[List[NoiseType], NoiseType] = NoiseType.ADDITIVE,
    name: Optional[str] = None,
) -> Problem:
    """
    Add additive or multiplicative noise drawn from a distribution in scipy.stats
    https://docs.scipy.org/doc/scipy/reference/stats.html
    Args:
        problem: problem instance where we want to add noise to
        random_variables: list of instances or instance of random variables, e.g.,
                          [scipy.stats.beta, scipy.stats.cauchy], each element of the
                          list is applied to the corresponding data column
        rv_kwargs: keyword arguments for the random variables, type must correspond to
                   the on of random_variables if given.
        noise_types: list of the same length as random_variables or single value,
                     decides whether the noise is added or multiplied
        name: name of the noisy problem

    Returns: noisy problem instance

    Example:
        To see how to obtain noisy flow-reactor-problems check the following examples.

        1.) Additive Gaussian noise for all output columns

        from scipy import stats
        from opti.problems.flow_reactor_unconstrained import get_problem_pareto
        flow = get_problem_pareto()
        rv = stats.norm
        noisy_flow = noisify_problem_with_scipy_stats(flow, rv)

        2.) Additive Gaussian noise for all output columns with non-standard parameters

        from scipy import stats
        from opti.problems.flow_reactor_unconstrained import get_problem_pareto
        flow = get_problem_pareto()
        rv = stats.norm
        rv_kwargs = {"loc": 0.5, "scale": 0.3}
        noisy_flow = noisify_problem_with_scipy_stats(flow, rv, rv_kwargs=rv_kwargs)

        3.) Different noise distributions and types for the different output columns

        from scipy import stats
        from opti.problems.flow_reactor_unconstrained import get_problem_pareto
        flow = get_problem_pareto()
        rvs = [stats.truncnorm, stats.norm, stats.gamma]
        rv_kwargs = [{"a": 0.1, "b": 2}, {"loc": 0.5, "scale": 0.3}, {"a": 1.99}]
        noise_types = [NoiseType.ADDITIVE, NoiseType.ADDITIVE, NoiseType.MULTIPLICATIVE]
        noisy_flow = noisify_problem_with_scipy_stats(
            flow, rvs, rv_kwargs=rv_kwargs, noise_types=noise_types
        )
    """
    # listify single element inputs
    if not isinstance(random_variables, list):
        random_variables = [random_variables] * problem.n_outputs
        rv_kwargs = [rv_kwargs] if rv_kwargs is not None else [{}]
        rv_kwargs = rv_kwargs * problem.n_outputs
    if not isinstance(noise_types, list):
        noise_types = [noise_types] * problem.n_outputs

    ops = {
        NoiseType.ADDITIVE: lambda x, y: x + y,
        NoiseType.MULTIPLICATIVE: lambda x, y: x * y,
    }
    op_names = {
        NoiseType.ADDITIVE: "add",
        NoiseType.MULTIPLICATIVE: "mul",
    }

    def apply_noise(data_column, noise_type, random_variable, rv_kwargs):
        sampled_noise = random_variable.rvs(size=data_column.size, **rv_kwargs)
        return ops[noise_type](data_column, sampled_noise)

    noisifiers = [
        partial(apply_noise, noise_type=nt, random_variable=rv, rv_kwargs=rvk)
        for nt, rv, rvk in zip(noise_types, random_variables, rv_kwargs)
    ]
    if name is None:
        rv_type_names = [
            f"{rv.name}_{op_names[nt]}" for rv, nt in zip(random_variables, noise_types)
        ]
        name = f"{problem.name} with rvs { ','.join(rv_type_names)}"
    return noisify_problem(problem, noisifiers, name=name)


def noisify_problem_with_gaussian(problem: Problem, mu: float = 0, sigma: float = 0.05):
    """
    Given an instance of a problem, this returns the problem with additive Gaussian noise
    Args:
        problem: problem instance where we add the noise
        mu: mean of the Gaussian noise to be added
        sigma: standard deviation of the Gaussian noise to be added

    Returns: input problem with additive Gaussian noise
    """
    rv_kwargs = {"loc": mu, "scale": sigma}
    noise_type = NoiseType.ADDITIVE
    return noisify_problem_with_scipy_stats(problem, stats.norm, rv_kwargs, noise_type)



import numpy as np

from opti.constraint import LinearInequality, NChooseK
from opti.objective import Maximize
from opti.parameter import Continuous, Parameters
from opti.problem import Problem


class Detergent(Problem):
    """Constrained problem with 5 inputs and 5 outputs.

    Each output is modeled as a second degree polynomial.

    The sixth input is a filler (water) and is factored out using the formulation
    constraint sum x = 1, and it's parameter bounds 0.6 < water < 0.8 result in 2 linear
    inequality constraints for the other parameters.
    """

    def __init__(self):
        # coefficients for the 2-order polynomial; generated with
        # base = 3 * np.ones((1, 5))
        # scale = PolynomialFeatures(degree=2).fit_transform(base).T
        # coef = np.random.RandomState(42).normal(scale=scale, size=(len(scale), 5))
        # coef = np.clip(coef, 0, None)
        coef = np.array(
            [
                [0.4967, 0.0, 0.6477, 1.523, 0.0],
                [0.0, 4.7376, 2.3023, 0.0, 1.6277],
                [0.0, 0.0, 0.7259, 0.0, 0.0],
                [0.0, 0.0, 0.9427, 0.0, 0.0],
                [4.3969, 0.0, 0.2026, 0.0, 0.0],
                [0.3328, 0.0, 1.1271, 0.0, 0.0],
                [0.0, 16.6705, 0.0, 0.0, 7.4029],
                [0.0, 1.8798, 0.0, 0.0, 1.7718],
                [6.6462, 1.5423, 0.0, 0.0, 0.0],
                [0.0, 0.0, 9.5141, 3.0926, 0.0],
                [2.9168, 0.0, 0.0, 5.5051, 9.279],
                [8.3815, 0.0, 0.0, 2.9814, 8.7799],
                [0.0, 0.0, 0.0, 0.0, 7.3127],
                [12.2062, 0.0, 9.0318, 3.2547, 0.0],
                [3.2526, 13.8423, 0.0, 14.0818, 0.0],
                [7.3971, 0.7834, 0.0, 0.8258, 0.0],
                [0.0, 3.214, 13.301, 0.0, 0.0],
                [0.0, 8.2386, 2.9588, 0.0, 4.6194],
                [0.8737, 8.7178, 0.0, 0.0, 0.0],
                [0.0, 2.6651, 2.3495, 0.046, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]
        )

        def poly2(x: np.ndarray) -> np.ndarray:
            """Full quadratic feature expansion including bias term."""
            return np.concatenate([[1], x, np.outer(x, x)[np.triu_indices(5)]])

        def f(x):
            x = np.atleast_2d(x)
            xp = np.stack([poly2(xi) for xi in x], axis=0)
            return xp @ coef

        inputs = Parameters(
            [
                Continuous("x1", domain=[0.0, 0.2]),
                Continuous("x2", domain=[0.0, 0.3]),
                Continuous("x3", domain=[0.02, 0.2]),
                Continuous("x4", domain=[0.0, 0.06]),
                Continuous("x5", domain=[0.0, 0.04]),
            ]
        )

        super().__init__(
            name="Detergent",
            inputs=inputs,
            outputs=[Continuous(f"y{i+1}", domain=[0, 3]) for i in range(5)],
            objectives=[Maximize(f"y{i+1}") for i in range(5)],
            constraints=[
                LinearInequality(names=inputs.names, lhs=-np.ones(5), rhs=-0.2),
                LinearInequality(names=inputs.names, lhs=np.ones(5), rhs=0.4),
            ],
            f=f,
        )


class Detergent_NChooseKConstraint(Problem):
    """Variant of the Detergent problem with an n-choose-k constraint"""

    def __init__(self):
        base = Detergent()

        super().__init__(
            name="Detergent with n-choose-k constraint",
            inputs=base.inputs,
            outputs=base.outputs,
            objectives=base.objectives,
            constraints=list(base.constraints)
            + [NChooseK(names=base.inputs.names, max_active=3)],
            f=base.f,
        )


class Detergent_OutputConstraint(Problem):
    """Variant of the Detergent problem with an output constraint.

    There are 5 inputs, 6 outputs:
    - 3 outputs are to be maximized
    - 1 output represents the stability of the formulation
        (0: not stable, 1: stable)
    """

    def __init__(self):
        base = Detergent()

        def f(x):
            x = np.atleast_2d(x)
            y1_5 = base.f(x)
            y6 = (0.4 - x.sum(axis=1)) / 0.2  # continuous version
            # y6 = np.sum(x, axis=1) < 0.3  # discrete version
            return np.column_stack([y1_5, y6])

        super().__init__(
            name="Detergent with stability constraint",
            inputs=base.inputs,
            outputs=list(base.outputs) + [Continuous("stable", domain=[0, 1])],
            objectives=[Maximize(f"y{i+1}") for i in range(3)],
            output_constraints=[Maximize("stable", target=0.5)],
            constraints=base.constraints,
            f=f,
        )



import io
import re

import numpy as np
import pandas as pd

import opti


class MipFile:
    """File reader for MODDE investigation files (.mip)"""

    def __init__(self, filename):
        """Read in a MODDE file.

        Args:
            filename (str or path): path to MODDE file
        """
        s = b"".join(open(filename, "rb").readlines())

        # Split into blocks of 1024 bytes
        num_blocks = len(s) // 1024
        assert len(s) % 1024 == 0
        blocks = [s[i * 1024 : (i + 1) * 1024] for i in range(num_blocks)]

        # Split blocks into header (first 21 bytes) and body
        headers = [b[0:21] for b in blocks]
        data = [b[21:] for b in blocks]

        # Parse the block headers to create a mapping {block_index: next_block_index}
        block_order = {}
        for i, header in enumerate(headers):
            i_next = np.where([h.startswith(header[15:19]) for h in headers])[0]
            block_order[i] = i_next[0] if len(i_next) == 1 else None

        # Join all blocks that belong together and decode to UTF-8
        self.parts = []
        while len(block_order) > 0:
            i = next(iter(block_order))  # get first key in ordered dict
            s = b""
            while i is not None:
                s += data[i]
                i = block_order.pop(i)
            s = re.sub(b"\r|\x00", b"", s).decode("utf-8", errors="ignore")
            self.parts.append(s)

        self.settings = self._get_design_settings()
        self.factors = {k: self.settings[k] for k in self.settings["Factors"]}
        self.responses = {k: self.settings[k] for k in self.settings["Responses"]}
        self.data = self._get_experimental_data()

    def _get_experimental_data(self):
        """Parse the experimental data.

        Returns:
            pd.DataFrame: dataframe of experimental data
        """
        part = [p for p in self.parts if p.startswith("ExpNo")][0]
        return pd.read_csv(io.StringIO(part), delimiter="\t", index_col="ExpNo")

    def _get_design_settings(self):
        """Parse the design settings.

        Returns:
            dict of dicts: dictionary with 'Factors', 'Responses', 'Options' as well as the individual variables.
        """
        part = [p for p in self.parts if p.startswith("[Status]")][0]
        settings = {}
        for line in part.split("\n"):
            if len(line) == 0:
                continue
            if line.startswith("["):
                thing = line.strip("[]")
                settings[thing] = {}
            else:
                key, value = line.split("=")
                settings[thing][key] = value
        return settings


def read_modde(filepath):
    """Read a problem specification from a MODDE file.

    Args:
        filepath (path-like): path to MODDE .mip file

    Returns:
        opti.Problem: problem specification
    """
    mip = MipFile(filepath)

    inputs = []
    outputs = []
    constraints = []
    formulation_parameters = []

    # build input space
    for name, props in mip.factors.items():
        if props["Use"] == "Uncontrolled":
            print(f"Uncontrolled factors not supported. Skipping {name}")
            continue
        if props["Type"] == "Formulation":
            domain = [float(s) for s in props["Settings"].split(",")]
            inputs.append(opti.Continuous(name=name, domain=domain))
            formulation_parameters.append(name)
        elif props["Type"] == "Quantitative":
            domain = [float(s) for s in props["Settings"].split(",")]
            inputs.append(opti.Continuous(name=name, domain=domain))
        elif props["Type"] == "Multilevel":
            domain = [float(s) for s in props["Settings"].split(",")]
            inputs.append(opti.Discrete(name=name, domain=domain))
        elif props["Type"] == "Qualitative":
            domain = props["Settings"].split(",")
            inputs.append(opti.Categorical(name=name, domain=domain))
    inputs = opti.Parameters(inputs)

    # build formulation constraint
    constraints.append(
        opti.constraint.LinearEquality(
            names=formulation_parameters,
            lhs=np.ones(len(formulation_parameters)),
            rhs=1,
        )
    )

    # build output space
    for name, props in mip.responses.items():
        # check if data available that allows to infer the domain
        vmin = mip.data[name].min()
        vmax = mip.data[name].max()
        if np.isfinite(vmin) or np.isfinite(vmax) and vmin < vmax:
            domain = [vmin, vmax]
        else:
            domain = [0, 1]
        dim = opti.Continuous(name=name, domain=domain)
        outputs.append(dim)
    outputs = opti.Parameters(outputs)

    # data
    data = mip.data.drop(columns=["ExpName", "InOut"])

    return opti.Problem(
        inputs=inputs, outputs=outputs, constraints=constraints, data=data
    )
